import pandas as pd
import os
import ast

def init_simulation_configs(path):
    """
    Loads the config file, reading the csv given in path, and adding their values to the global scope
    Returns the names of the values in a list
    Input:
        string: path
        return: array of all the names of the variables
    """ 
    df_configs = pd.read_csv(path, header=0, converters={"Values": ast.literal_eval})
    if len(df_configs) < 31:
        raise Exception("Less than 31 configuration options! Are there some missing?")
    if len(df_configs) > 31:
        raise Exception("More than 31 configuration options!")
    dict_configs = dict(zip(df_configs["Names"], df_configs["Values"]))
    globals().update(dict_configs)  
    error_string = ""
    if sum(extravasation_probs) != 1:
        error_string += "Extravasation probabilities must sum 1!\n"
    if len(extravasation_probs) != grids_number - 1:
        error_string += "There must be as many Extravasation probabilities as the value of (grids_number - 1)!\n"
    if len(secondary_sites_vessels) != grids_number-1:
        error_string += "There must be as many secondary site vessels as the value of grids_number - 1!\n"
    if mesenchymal_proportion + epithelial_proportion != 1:
        error_string += "Mesenchymal_proportion + epithelial_proportion must be 1!\n"
    if error_string != "":
        raise ValueError(error_string)
    return(list(df_configs["Names"]))

def load_simulation_configs_for_data_generation(path):
    """
    Loads the config file, reading the csv given in path, and adding their values to the global scope
    Returns the names of the values in a list
    Input:
        string: path
        return: array of all the names of the variables
    """ 
    df_configs = pd.read_csv(path, header=0, converters={"Values": ast.literal_eval})
    if len(df_configs) < 33:
        raise Exception("Less than 33 configuration options! Are there some missing?")
    if len(df_configs) > 33:
        raise Exception("More than 33 configuration options!")
    dict_configs = dict(zip(df_configs["Names"], df_configs["Values"]))
    globals().update(dict_configs)  
    return(list(df_configs["Names"]))

def load_simulation_configs_for_reloaded_simulation(path):
    """
    Loads the config file, reading the csv given in path, and adding their values to the global scope
    Returns the names of the values in a list
    Input:
        string: path
        return: array of all the names of the variables
    """ 
    df_configs = pd.read_csv(path, header=0, converters={"Values": ast.literal_eval})
    # if the configs were loaded to continue from a previous 
    # simulation, drop the last two rows
    # they shoud not be loaded as they change for each simulation according to user input
    # (max steps and step size)
    df_configs = df_configs[:-2]
    if len(df_configs) < 31:
        raise Exception("Less than 33 configuration options! Are there some missing?")
    if len(df_configs) > 31:
        raise Exception("More than 33 configuration options!")
    dict_configs = dict(zip(df_configs["Names"], df_configs["Values"]))
    globals().update(dict_configs)  
    return(list(df_configs["Names"]))

def generate_default_configs():
    """Creates a default simulation_configs.csv file"""
    names = ["th","tha","xh","xha","dM","dE","phiM","phiE","dmmp","theta","Lambda","gamma1","gamma2","vasculature_time","doubling_time_M","doubling_time_E","single_cell_survival","cluster_survival","extravasation_probs","dissagreggation_prob","carrying_capacity","normal_vessels_primary","ruptured_vessels_primary","secondary_sites_vessels","n_center_points_for_tumor","n_center_points_for_Vessels","gridsize","grids_number","mesenchymal_proportion","epithelial_proportion","number_of_initial_cells"]
    values = [0.001,0.001,0.005,0.005,1e-4,5e-5,0.0005,0.0005,0.001,0.195,0.1,1,1,180,2000,3000,5e-04,0.025,[0.75, 0.25],0.5,4,8,2,[10, 10],97,200,201,3,0.6,0.4,388]
    default_configs = pd.DataFrame({"Names": names, "Values": values})
    default_configs.to_csv("simulations_configs.csv", index=False)

def check_if_configs_are_present():
    """Checks if a configs file is present in the current directory"""
    if not os.path.isfile("simulations_configs.csv"):
        print("Configs file`for new simulation \"simulations_configs.csv\" not found! Creating a configs file with default values.")
        generate_default_configs()


import ast
import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
import re
import os
import sys
import metaspread.configs

def get_equally_spaced_array(passed_array, number_of_elems):
    passed_array = np.array(passed_array)
    indexes = np.round(np.linspace(0, len(passed_array)-1, number_of_elems)).astype(int)
    return list(zip(indexes,passed_array[indexes]))

def plot_cancer(fig_counter, grid_id, step, real_time_at_step, simulation_path, tumor_images_path):
    plt.style.use("seaborn-v0_8-darkgrid")
    figure_path = os.path.join(tumor_images_path, f'Cells-grid{grid_id}-step{step} - Tumor size at {real_time_at_step/(3600*24):.2f} days.png')
    if os.path.isfile(figure_path):
        return
    
    all_coords_data_path = os.path.join(simulation_path, "Data analysis", "Tumor dynamics")
    current_coords_filename = [file for file in os.listdir(all_coords_data_path) if file.startswith(f"Cells-grid{grid_id}-step{step} - Tumor size at")][0]
    current_coords_path = os.path.join(all_coords_data_path, current_coords_filename)
    coords_list = pd.read_csv(current_coords_path, index_col=0)
    Xm, Ym, Xe, Ye, Xv, Yv, Xvr, Yvr = coords_list.iloc[0], coords_list.iloc[1], coords_list.iloc[2], coords_list.iloc[3], coords_list.iloc[4], coords_list.iloc[5], coords_list.iloc[6], coords_list.iloc[7]
    plt.figure(fig_counter, figsize=(6, 6), facecolor='white')
    
    plt.scatter(Xm, Ym, marker='o', color='blue', alpha=0.5/metaspread.configs.carrying_capacity, label="Mesenchymal cells")
    plt.scatter(Xe, Ye, marker='h', color='orange', alpha=0.5/metaspread.configs.carrying_capacity, label="Epithelial cells")
    plt.scatter(Xv, Yv, marker='.', color='red', alpha=0.8, label="Vasculature points")
    plt.scatter(Xvr, Yvr, marker='+', color='darkred', alpha=0.8, label="Ruptured vasculature points")
    plt.xlim(0, metaspread.configs.gridsize)
    plt.ylim(0, metaspread.configs.gridsize)

    xticks = np.arange(0, metaspread.configs.gridsize, step=int(metaspread.configs.gridsize/6)) # 6 ticks
    xticklabels = [str(round(j,1)) for j in np.arange(0, 2.1, step = 2/201*(metaspread.configs.gridsize/6))]
    plt.xticks(xticks, xticklabels)
    plt.yticks(xticks, xticklabels)
    plt.xlabel("mm")
    plt.ylabel("mm")
    plt.grid(False)

    if step <= 2000:
        plt.legend(loc="upper left")
    plt.title(f'Tumor size at {real_time_at_step/(3600*24):.2f} days ({step} steps) - grid {grid_id}', fontsize = 13)

    # save the figure
    plt.savefig(figure_path)


def plot_growth_data(simulation_path, cells_images_path, grid_id, step, real_time_at_step):
    plt.style.use("seaborn-v0_8-darkgrid")
    path_to_save = os.path.join(cells_images_path, f'CellsGrowth-grid{grid_id}-step{step} - {real_time_at_step/(3600*24):.2f} days.png')
    if os.path.isfile(path_to_save):
        return
    all_cells_growth_data_path = os.path.join(simulation_path, "Data analysis", "Cells growth")
    current_growth_filename = [file for file in os.listdir(all_cells_growth_data_path) if file.startswith(f"CellsGrowth-grid{grid_id}-step{step} - ")][0]
    current_growth_path = os.path.join(all_cells_growth_data_path, current_growth_filename)
    df_cells_number = pd.read_csv(current_growth_path, index_col=0)

    plt.plot(df_cells_number["Days"], df_cells_number["Number of Mesenchymal Cells"], label="Mesenchymal cells", color='tab:blue')
    plt.plot(df_cells_number["Days"], df_cells_number["Number of Epithelial Cells"], label="Epithelial cells", color='tab:orange')

    plt.title(f'Cells growth - grid {grid_id}', fontsize = 13)
    plt.xlabel('Days')
    plt.ylabel('Number of cells')
    plt.legend(loc="upper left")
    plt.ylim(0)

    # save the figure
    plt.savefig(path_to_save)

def plot_MMP2_or_ECM(i, step, real_time_at_step, files_path, fig_counter, grid_id, path_to_save, type="Mmp2"):
    if type=="Mmp2":
        figure_path = os.path.join(path_to_save, f'{type}-grid{grid_id}-step{step} - {real_time_at_step/(3600*24):.2f} days.png')
    elif type=="Ecm":
        figure_path = os.path.join(path_to_save, f'{type}-grid{grid_id}-step{step} - {real_time_at_step/(3600*24):.2f} days.png')
    if os.path.isfile(figure_path):
        return
    try:
        df = pd.read_csv(files_path[i], index_col=0)
    except:
        raise Exception(f"Corrupted or empty file {files_path[i]}")
    plt.figure(fig_counter, figsize=(6, 5), facecolor='white')

    if type=="Mmp2":
        plt.imshow(df.T, vmin=0, vmax=3)
    elif type == "Ecm":
        plt.imshow(df.T , vmin=0, vmax=1)
        
    plt.colorbar()

    plt.xlim(0, metaspread.configs.gridsize)
    plt.ylim(0, metaspread.configs.gridsize)

    xticks = np.arange(0, metaspread.configs.gridsize, step=int(metaspread.configs.gridsize/6)) # 6 ticks
    xticklabels = [str(round(j,1)) for j in np.arange(0, 2.1, step = 2/201*(metaspread.configs.gridsize/6))]
    plt.xticks(xticks, xticklabels)
    plt.yticks(xticks, xticklabels)
    plt.xlabel("mm")
    plt.ylabel("mm")
    plt.grid(visible=None)

    plt.title(f'{type} at {real_time_at_step/(3600*24):.2f} days ({step} steps) - grid {grid_id}', fontsize = 13)
    
    plt.savefig(figure_path)

def plot_histogram(histogram_csv_file_path, all_histogram_images_path, step, real_time_at_step, grid_id):
    plt.style.use("seaborn-v0_8-darkgrid")
    path_to_save = os.path.join(all_histogram_images_path, f"Cells-grid{grid_id}-step{step} - Histogram at {real_time_at_step/(3600*24):.2f} days.png")
    if os.path.isfile(path_to_save):
        return
    histogram = pd.read_csv(histogram_csv_file_path, header=0, index_col=0)
    if histogram.empty:
        return
    plt.xlabel('Number of cells per grid-point')
    plt.ylabel('\nNr. grid-points with given nr. of cells')
    plt.title(f'Positions histogram at {real_time_at_step/(3600*24):.2f} days ({step} steps) - grid {grid_id}', fontsize = 13)
    plt.bar(histogram['Bins'], histogram['Frequency'])
    plt.xticks(range(metaspread.configs.carrying_capacity + 1))
    plt.xlim([-1, metaspread.configs.carrying_capacity + 1])
    plt.savefig(path_to_save)

def plot_vasculature_graphs(vasculature_df, path_to_save, max_step, real_delta_time):
    plt.style.use("seaborn-v0_8-darkgrid")
    figure_path = os.path.join(path_to_save, f'Vasculature-cells-step{max_step}.png')
    if not os.path.isfile(figure_path):
        # Prepare the data for the bar chart
        mesenchymal_data = vasculature_df["Mesenchymal cells"]
        epithelial_data = vasculature_df["Epithelial cells"]
        time_steps = vasculature_df["Time"]
        plt.style.use("seaborn-v0_8-darkgrid")

        #second plot, clusters
        plt.figure(facecolor='white')
        
        # Plot the data for each category
        plt.step(time_steps, mesenchymal_data, where='mid', color='tab:blue', label='Mesenchymal Cells', linewidth=1.5)
        plt.step(time_steps, epithelial_data, where='mid', color='tab:orange', label='Epithelial Cells', linewidth=1.5)

        # Set the chart labels, title, and legend
        plt.xlabel('Day')
        plt.ylabel('Cell Count')
        plt.title('Vasculature Cell Counts')
        plt.xticks([0, max_step/2, max_step], [0, f"{(max_step/2)*real_delta_time/(3600*24):.2f}", f"{max_step*real_delta_time/(3600*24):.2f}"])
        
        plt.xlim([-0.5, max_step + 0.5])
        plt.legend(loc="upper left")

        # save the figure
        plt.savefig(figure_path)
        plt.close()

    figure_path = os.path.join(path_to_save, f'Vasculature-clusters-step{max_step}.png')
    if not os.path.isfile(figure_path):
        total_cluster_data = vasculature_df["Total clusters"]
        multicellular_cluster_data = vasculature_df["Multicellular clusters"]
        time_steps = vasculature_df["Time"]
        #second plot, clusters
        plt.figure(facecolor='white')
        
        # Plot the data for each category
        plt.step(time_steps, total_cluster_data, where='mid', color='darkred', label='Total clusters', linewidth=1.5)
        plt.step(time_steps, multicellular_cluster_data, where='mid', color='green', label='Multicellular clusters', linewidth=1.5)

        # Set the chart labels, title, and legend
        plt.xlabel('Day')
        plt.ylabel('Cluster Count')
        plt.title('Vasculature Cluster Counts')
        plt.xticks([0, max_step/2, max_step], [0, f"{(max_step/2)*real_delta_time/(3600*24):.2f}", f"{max_step*real_delta_time/(3600*24):.2f}"])
        
        plt.xlim([-0.5, max_step + 0.5])
        plt.legend(loc="upper left")

        # save the figure
        plt.savefig(figure_path)
        plt.close()

def plot_radius_diameter_history(df, path_to_save, max_step, real_delta_time):
    plt.style.use("seaborn-v0_8-darkgrid")
    figure_path = os.path.join(path_to_save, 'Radius and diameter for Grid 1.png')
    if not os.path.isfile(figure_path):
        # Prepare the data for the bar chart
        radius = df["Radius"]*metaspread.configs.xh*(0.001/0.005)*10  #getting distance in mm
        diameter = df["Diameter"]*metaspread.configs.xh*(0.001/0.005)*10 #getting distance in mm
        time_steps = df["Step"]
        plt.style.use("seaborn-v0_8-darkgrid")

        #second plot, clusters
        plt.figure(facecolor='white')
        
        # Plot the data for each category
        plt.step(time_steps, radius, where='mid', color='tab:blue', label='Radius', linewidth=1.5)
        plt.step(time_steps, diameter, where='mid', color='tab:orange', label='Diameter', linewidth=1.5)

        # Set the chart labels, title, and legend
        plt.xlabel('Day')
        plt.ylabel('Distance (mm)')
        plt.title('Radius and diameter for Grid 1')
        plt.xticks([0, max_step/2, max_step], [0, f"{(max_step/2)*real_delta_time/(3600*24):.2f}", f"{max_step*real_delta_time/(3600*24):.2f}"])
        
        plt.xlim([-0.5, max_step + 0.5])
        plt.legend(loc="upper left")

        # save the figure
        plt.savefig(figure_path)
        plt.close()

    
def generate_graphs(name_of_the_simulation, amount_of_pictures=0):
    simulations_dir = "Simulations"
    simulation_path = os.path.join(simulations_dir, name_of_the_simulation)
    configs_path = os.path.join(simulation_path, "configs.csv")
    metaspread.configs.load_simulation_configs_for_data_generation(configs_path)
    print(f'Analyzing data in the folder {simulation_path}\n')

    # Get the Ecm and Mmp2 data filenames 
    ecm_path = os.path.join(simulation_path, "Ecm")
    mmp2_path = os.path.join(simulation_path, "Mmp2")
    ecm_files_name = [f for f in sorted(os.listdir(ecm_path), key=lambda x: int(re.findall(r'\d+(?=step)', x)[0])) if os.path.isfile(os.path.join(ecm_path, f)) and f.endswith(".csv")]
    mmp2_files_name = [f for f in sorted(os.listdir(mmp2_path), key=lambda x: int(re.findall(r'\d+(?=step)', x)[0])) if os.path.isfile(os.path.join(mmp2_path, f)) and f.endswith(".csv")]

    # Get the vasculature data filename
    vasculature_path = os.path.join(simulation_path, "Vasculature")
    vasculature_files_name = [f for f in os.listdir(vasculature_path) if os.path.isfile(os.path.join(vasculature_path, f)) and f.endswith(".json")]

    all_cells_filename = os.path.join(simulation_path, "CellsData.csv")
    if os.path.isfile(all_cells_filename):
        print("Using cells data at:", all_cells_filename)
    else:
        print("No CellsData.csv cell data found in directory:", simulation_path)
        return

    if ecm_files_name:
        ecm_files_path = [os.path.join(ecm_path, p) for p in ecm_files_name]
        print("Using Ecm data in the folder:", ecm_path)
    else:
        print("No .csv Ecm data found in directory:", ecm_path)
        return

    if mmp2_files_name:
        mmp2_files_path = [os.path.join(mmp2_path, p) for p in mmp2_files_name]
        print("Using Mmp2 data in the folder:", mmp2_path)
    else:
        print("No .csv Mmp2 data found in directory:", mmp2_path)
        return

    if vasculature_files_name:
        print("Using vasculature data at:", vasculature_path)
    else:
        print("No .json vasculature data found in directory:", simulation_path)
        return


    #Path of the data
    data_folder = "Data analysis"
    data_path = os.path.join(simulation_path, data_folder)

    tumor_data_path = os.path.join(data_path, "Tumor dynamics")
    
    step_size = metaspread.configs.data_collection_period
    real_delta_time = 40 * metaspread.configs.th/0.001 #in seconds (the original ratio is 40 seconds/0.001 non-dimensional time)
    grids_number = metaspread.configs.grids_number
    configs_max_step = metaspread.configs.max_steps
    all_cells_dataframe = pd.read_csv(all_cells_filename, converters={"Position": ast.literal_eval})
    all_cells_dataframe = all_cells_dataframe[["Step", "Position", "Phenotype", "Grid", "Agent Type", "Ruptured"]]
    max_step = max(all_cells_dataframe["Step"])
    if configs_max_step >= max_step:
        print(f"Warning: the run for this simulation terminated early")
        print(f"Max step reached is {max_step} while {configs_max_step} was expected.")
    maximum_frames = int(max_step / step_size)
    if amount_of_pictures > maximum_frames:
        print(f"There are only {maximum_frames} frames available!")
        print("Setting up the amount of picture to this value.")
        amount_of_pictures = 0 #to print all the pictures
    
    # Path to save all the images:
    images_folder = "Graphical analysis"
    images_path = os.path.join(simulation_path, images_folder)

    tumor_images_path = os.path.join(images_path, "Tumor dynamics")
    cells_images_path = os.path.join(images_path, "Cells growth")
    ecm_images_path = os.path.join(images_path, "Ecm dynamics")
    mmp2_images_path = os.path.join(images_path, "Mmp2 dynamics")
    vasculature_images_path = os.path.join(images_path, "Vasculature dynamics")
    all_histogram_images_path = os.path.join(images_path, "Positions histogram")
    radius_diameter_images_path = os.path.join(images_path, "Radius and diameter")

    # Create folder for all the Graphical analysis

    os.makedirs(images_path, exist_ok = True)
    os.makedirs(tumor_images_path, exist_ok = True)
    os.makedirs(mmp2_images_path, exist_ok = True)
    os.makedirs(ecm_images_path, exist_ok = True)
    os.makedirs(cells_images_path, exist_ok = True)
    os.makedirs(vasculature_images_path, exist_ok = True)
    os.makedirs(all_histogram_images_path, exist_ok = True)
    os.makedirs(radius_diameter_images_path, exist_ok = True)

    print(f"\nSaving tumor images in the folder:", tumor_images_path)
    print("Saving Mmp2 images in the folder:", mmp2_images_path)
    print("Saving Ecm images in the folder:", ecm_images_path)
    print("Saving cells numbers images in the folder:", cells_images_path)
    print("Saving vasculature images in the folder:", vasculature_images_path)
    print("Saving histogram images in the folder:", all_histogram_images_path)
    print("Saving histogram image in the folder:", radius_diameter_images_path)

    
    if amount_of_pictures != 0:
        range_of_pictures = get_equally_spaced_array(range(step_size,max_step+1,step_size), amount_of_pictures)
    else:
        range_of_pictures = enumerate(range(step_size,max_step+1,step_size))
    fig_counter = 1
    for grid_id in range(1, grids_number+1):
        plt.style.use("default")
        plt.style.use("Solarize_Light2")
        print(f'\nGrid: {grid_id}')
    
        # Plot the Mmp2 graphs
        print(f'\tPlotting Mmp2 graphs...')
        for id, step in range_of_pictures:
            real_time_at_step = real_delta_time * step
            mmp2_files_path_this_grid = [path for path in mmp2_files_path if f"Mmp2-{grid_id}grid-" in path]
            plot_MMP2_or_ECM(id, step, real_time_at_step, mmp2_files_path_this_grid, fig_counter, grid_id, mmp2_images_path, type="Mmp2")
            plt.close()
            fig_counter += 1

        # Plot the Ecm graphs
        print(f'\tPlotting Ecm graphs...')
        for id, step in range_of_pictures:
            real_time_at_step = real_delta_time * step
            ecm_files_path_this_grid = [path for path in ecm_files_path if f"Ecm-{grid_id}grid-" in path]
            plot_MMP2_or_ECM(id, step, real_time_at_step, ecm_files_path_this_grid, fig_counter, grid_id, ecm_images_path, type="Ecm")
            plt.close()
            fig_counter += 1

        # Plot the cells graphs
        print(f'\tPlotting tumor graphs...')
        for id, step in range_of_pictures:
            real_time_at_step = real_delta_time * step
            plot_cancer(fig_counter, grid_id, step, real_time_at_step, simulation_path, tumor_images_path)
            plt.close()
            fig_counter += 1

        # Plot the histogram graphs
        print(f'\tPlotting histogram graphs...')
        csv_histogram_files_names = [f for f in os.listdir(tumor_data_path) if os.path.isfile(os.path.join(tumor_data_path, f)) and "Histogram" in f]
        for id, step in range_of_pictures:
            real_time_at_step = real_delta_time * step
            histogram_csv_file_name = csv_histogram_files_names[id]
            histogram_csv_file_path = os.path.join(tumor_data_path, histogram_csv_file_name)
            plot_histogram(histogram_csv_file_path, all_histogram_images_path, step, real_time_at_step, grid_id)
            plt.close()
            fig_counter += 1

        # Plot the growth of ephitelial and mesenchymal cells
        print(f'\tPlotting cells numbers graph...')
        for id, step in range_of_pictures:
            real_time_at_step = real_delta_time * step
            plot_growth_data(simulation_path, cells_images_path, grid_id, step, real_time_at_step)
            plt.close()
            fig_counter += 1

    # Plot the vasculature data
    print(f'Plotting vasculature...')
    for id, step in range_of_pictures:
        folder_path = os.path.join(simulation_path, "Data analysis", "Vasculature dynamics")
        try:
            file_path = os.path.join(folder_path, f"Vasculature-step{step}.csv")
            vasculature_data = pd.read_csv(file_path, header=0)
        except:
            print(f"Error while reading the vasculature data for step {step} in grid {grid_id}", file=sys.stderr)
            print("Did you run the 'Data analysis' in the postprocessing menu first?", file=sys.stderr)
            os._exit(1)
        plot_vasculature_graphs(vasculature_data, vasculature_images_path, step, real_delta_time)
        plt.close()
        fig_counter += 1
        
    #plotting the radius and diameter history graph
    print(f'Plotting radius and diameter history graph...')
    folder_path = os.path.join(simulation_path, "Data analysis", "Tumor dynamics")
    file_name = [file for file in os.listdir(folder_path) if file.startswith("Tumor radius and diameter history")][0]
    file_path = os.path.join(folder_path, file_name)
    radius_history_df = pd.read_csv(file_path)#, header=0)
    plot_radius_diameter_history(radius_history_df, radius_diameter_images_path, step, real_delta_time)
    plt.close()
    fig_counter += 1

    plt.show()
    plt.close()

    plt.show()
    plt.close()









import mesa
import matplotlib.pyplot as plt
import numpy as np
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
import pandas as pd
import os
import json
import ast
from metaspread.cancercell import CancerCell
from metaspread.vessel import Vessel
from metaspread.quasicircle import find_quasi_circle
from matplotlib import pyplot as plt
from matplotlib import cm
# from Classes.configs import *
import metaspread.configs
# import pickle


def get_cluster_survival_probability(cluster):
    """
    Takes in a tuple representing a cluster, returns the survival probabiltiy.
    
    Input:
        Cluster: a tuple representing the cancer cells cluster, where the first 
        element corresponds to the amount of Mesenchymal cells, and the second
        corresponds to the amount of Epithelial cells.
    Returns:
        The corresponding survival probability, according to if its a songle-cell
        cluster or a multi-cellular one.
    """
    if cluster[0] < 0:
        raise Exception(f"Error! Mesenchymal cells are negative: {cluster[0]}")
    if cluster[1] < 0:
        raise Exception(f"Error! Epithelial cells are negative: {cluster[1]}")
    if sum(cluster) == 1:
        return (single_cell_survival)
    elif sum(cluster) > 1:
        return (cluster_survival)
    elif sum(cluster) == 0:
        raise Exception(f"Error, no cells in cluster!")
    else:
        raise Exception(f"Error, nothing returned for cluster survival probability" )
    

def count_total_cells(model):
    """Counts all the cells present in the model, in all sites.

    Input:
        model: CancerModel object.
    Returns:
        amount_of_cells (int): the total amount of cells in every site,
        NOT considering the vasculature    
    """
    return sum(model.cancer_cells_counter)

def count_vasculature_cells(model):
    """"
    Counts the total amount of cells in the vasculature

    Input: 
        model: CancerModel object.
    Returns:
        amount_of_cells (int): the total amount of cells in the vasculature
    """
    amount_of_cells = sum([sum(x+y for x,y in value) for value in model.vasculature.values()])
    return amount_of_cells

class CancerModel(mesa.Model):
    """
    Class for the model.

    Attributes:
    ---------------
    number_of_initial_cells: int
        number of initial cancer cells
    width: int
        the width of each grid
    height: int
        the height of each grid
    grids_number: int
        the amount of sites, consideiring the initial tumour site + the secondary sites
    seed: int
        the seed used for the random number generation for all the simulation.
        If None, the random one will be selected by default.

    Methods:
    ---------------
    _initialize_grids__()
        initialize the grid with the initial bessel and cancer cell population
    proliferate(cell_type)
        Duplicates every cancer cell in the model of the cell_type phenotype
    calculate_environments(mmp2, ecm)
        Calculates the next step for the given arrays of mmp2 and ecm concentrations
    disaggregate_clusters(time)
        For a given time, it will dissagregate single cells from clusters
    """

    def __init__(self, number_of_initial_cells, width, height, grids_number, max_steps, data_collection_period, new_simulation_folder, loaded_simulation_path="", fixed_p_left=None, fixed_p_right=None, fixed_p_top=None, fixed_p_bottom=None, seed=None):
        super().__init__()  
        # self.simulations_dir = "Simulations"
        
        self.fixed_p_left=fixed_p_left
        self.fixed_p_right=fixed_p_right
        self.fixed_p_top=fixed_p_top
        self.fixed_p_bottom=fixed_p_bottom
        self.vasculature = {}
        self.number_of_initial_cells = number_of_initial_cells
        self.width = width
        self.height = height
        self.phenotypes = ["mesenchymal", "epithelial"]
        self.grid_vessels_positions = [[] for _ in range(grids_number)]
        self.current_agent_id = 0
        self.max_steps = max_steps
        self.data_collection_period = data_collection_period
        self.new_simulation_folder  = new_simulation_folder
        self.mesenchymal_count = [np.zeros((width, height), dtype=float) for _ in range(grids_number)]
        self.epithelial_count = [np.zeros((width, height), dtype=float) for _ in range(grids_number)]
        self.grids_number = grids_number
        self.grids = [mesa.space.MultiGrid(width, height, False) for _ in range(self.grids_number)]
        self.grid_ids = [i+1 for i in range(self.grids_number)]
        self.cancer_cells_counter = [0] * grids_number
        self.time_grid_got_populated = [-1 for _ in range(self.grids_number)]
        self.schedule = mesa.time.RandomActivation(self)
        #list of numpy arrays, representing mmp2 and ecm concentration in each grid
        self.mmp2 = [np.zeros((2, width, height), dtype=float) for _ in range(grids_number)]
        self.ecm = [np.ones((2, width, height), dtype=float) for _ in range(grids_number)]
        self.loaded_max_step = 0
        self.previous_cell_data = pd.DataFrame()

        if loaded_simulation_path != "":
            print(f"Loading simulation at {loaded_simulation_path}!")
            configs_path = os.path.join(loaded_simulation_path, "configs.csv")
            config_var_names = metaspread.configs.load_simulation_configs_for_reloaded_simulation(configs_path)
            #load the configs so we can use them in this module as globals
            for var in config_var_names:
                globals()[var] = getattr(metaspread.configs, var)
            self.load_previous_simulation(loaded_simulation_path)
            self.previous_cell_data = pd.read_csv(os.path.join(loaded_simulation_path, "CellsData.csv"), index_col=0)
        else:
            print("Starting simulation from zero!")
            configs_path = "simulations_configs.csv"
            config_var_names = metaspread.configs.init_simulation_configs(configs_path)
            #load the configs so we can use them in this module as globals
            for var in config_var_names:
                globals()[var] = getattr(metaspread.configs, var)
            self._initialize_grids()
            self.doubling_time_counter_M = doubling_time_M
            self.doubling_time_counter_E = doubling_time_E
        self.datacollector = mesa.DataCollector(
            model_reporters={"Total cells": count_total_cells}, agent_reporters={"Position": "pos", "Agent Type": "agent_type", "Phenotype": "phenotype", "Ruptured": "ruptured", "Grid": "grid_id"})

    def step(self):
        """Advance the model by one step.
        
        The step function of the model will be called on each of the siulations steps.
        It will correctly allocate the cells coming from the vaculature, if any.
        Then, it will calculate the ECM and MMP2 concentration changes for this step,
        proliferate the cells if its due, and collect the data if it is the appropiate time.

        Input: none
        Returns: none
        """       
        if self.schedule.time in self.vasculature: # Add keys
            self.disaggregate_clusters(self.schedule.time)
            surviving_clusters = [cluster for cluster in self.vasculature[self.schedule.time] if self.random.random() < get_cluster_survival_probability(cluster)]
            del self.vasculature[self.schedule.time]
            for cluster in surviving_clusters:
                selected_site = self.random.choices(range(1,self.grids_number), weights=extravasation_probs[0:self.grids_number-1])[0]
                arriving_point = self.random.choice(self.grid_vessels_positions[selected_site])
                x,y = arriving_point
                on_left_border    = self.grids[selected_site].out_of_bounds((x-1,y))
                on_right_border   = self.grids[selected_site].out_of_bounds((x+1,y))
                on_top_border     = self.grids[selected_site].out_of_bounds((x,y+1))
                on_bottom_border  = self.grids[selected_site].out_of_bounds((x,y-1))
                possible_places = self.grids[selected_site].get_neighborhood(arriving_point, moore=False, include_center=False)
                number_of_ccells_in_arriving_point ={}
                for x2,y2 in possible_places:
                    number_of_ccells_in_arriving_point[x2,y2] = len([agent for agent in self.grids[selected_site].get_cell_list_contents([(x2,y2)]) if agent.agent_type == "cell"])
                for tuple_index, ccells_amount in enumerate(cluster):
                    cell_type = "mesenchymal" if tuple_index == 0 else "epithelial"
                    while ccells_amount > 0:
                        if not on_left_border and carrying_capacity > number_of_ccells_in_arriving_point[x-1,y]:
                            ccell = CancerCell(self.current_agent_id, self, self.grids[selected_site], self.grid_ids[selected_site], cell_type, self.ecm[selected_site], self.mmp2[selected_site])
                            self.current_agent_id += 1
                            self.grids[selected_site].place_agent(ccell, (x-1,y)) 
                            number_of_ccells_in_arriving_point[x-1,y] += 1
                            self.cancer_cells_counter[selected_site] += 1
                            self.schedule.add(ccell)
                        elif not on_right_border and carrying_capacity > number_of_ccells_in_arriving_point[x+1,y]:
                            ccell = CancerCell(self.current_agent_id, self, self.grids[selected_site], self.grid_ids[selected_site], cell_type, self.ecm[selected_site], self.mmp2[selected_site])
                            self.current_agent_id += 1
                            self.grids[selected_site].place_agent(ccell, (x+1,y))
                            number_of_ccells_in_arriving_point[x+1,y] += 1
                            self.cancer_cells_counter[selected_site] += 1
                            self.schedule.add(ccell)
                        elif not on_bottom_border and carrying_capacity > number_of_ccells_in_arriving_point[x,y-1]:
                            ccell = CancerCell(self.current_agent_id, self, self.grids[selected_site], self.grid_ids[selected_site], cell_type, self.ecm[selected_site], self.mmp2[selected_site])
                            self.current_agent_id += 1
                            self.grids[selected_site].place_agent(ccell, (x,y-1))
                            number_of_ccells_in_arriving_point[x,y-1] += 1
                            self.cancer_cells_counter[selected_site] += 1
                            self.schedule.add(ccell)
                        elif not on_top_border and carrying_capacity > number_of_ccells_in_arriving_point[x,y+1]:
                            ccell = CancerCell(self.current_agent_id, self, self.grids[selected_site], self.grid_ids[selected_site], cell_type, self.ecm[selected_site], self.mmp2[selected_site])
                            self.current_agent_id += 1
                            self.grids[selected_site].place_agent(ccell, (x,y+1))
                            number_of_ccells_in_arriving_point[x,y+1] += 1
                            self.cancer_cells_counter[selected_site] += 1
                            self.schedule.add(ccell)
                        ccells_amount -= 1
                    
        #Perform ECM and MMP2 calculations
        self.calculate_environment(self.mmp2, self.ecm)
        
        # Proliferation
        # Counters are used so when loading a simulation the behaviour does not change, compared to use self.schedule.time % doubling_time_M == 0
        if (self.doubling_time_counter_M == 0 and self.schedule.time != 0):
            self.proliferate("mesenchymal")
            self.doubling_time_counter_M = doubling_time_M

        if (self.doubling_time_counter_E == 0 and self.schedule.time != 0):
            self.proliferate("epithelial")
            self.doubling_time_counter_E = doubling_time_E
                
        self.doubling_time_counter_E -= 1
        self.doubling_time_counter_M -= 1

        print(f'Step number: {self.schedule.time + self.loaded_max_step}', end="")
        print("\r", end="")
        self.schedule.step()
        
        #At the end of each step, check if the grid has been populated, and if it happened, store the time step when it did
        for index, time in enumerate(self.time_grid_got_populated):
            if time == -1: #if it has not been populated already, we check:
                if self.cancer_cells_counter[index] > 0:
                    self.time_grid_got_populated[index] = self.schedule.time + self.loaded_max_step

        # Saving of non agents data every data_collection_interval steps
        if (self.schedule.time != 0 and (self.schedule.time % self.data_collection_period == 0)) \
            or self.schedule.time == self.max_steps:
            self.datacollector.collect(self)
            current_agents_state = self.datacollector.get_agent_vars_dataframe()
            current_agents_state = current_agents_state.reset_index(level=["Step", "AgentID"])
            path_to_save = os.path.join(self.new_simulation_folder, f'CellsData.csv')
            if not self.previous_cell_data.empty:
                current_agents_state["Step"] += self.loaded_max_step
                current_agents_state = pd.concat([self.previous_cell_data, current_agents_state])
            current_agents_state.to_csv(path_to_save)
            #pickling a model could be an option in the future
            # backup_file_path = os.path.join(self.new_simulation_folder, "Backup", "backup.p")
            # with open(backup_file_path, "wb") as f:
            #     pickle.dump(self, f)
            df_time_grids_got_populated = pd.DataFrame()
            for grid_id in self.grid_ids:
                new_mmp2_df = pd.DataFrame(self.mmp2[grid_id-1][0,:,:])
                mmp2CsvName = f"Mmp2-{grid_id}grid-{self.schedule.time + self.loaded_max_step}step.csv"
                path_to_save = os.path.join(self.new_simulation_folder, "Mmp2", mmp2CsvName)
                new_mmp2_df.to_csv(path_to_save)

                new_ecm_df = pd.DataFrame(self.ecm[grid_id-1][0,:,:])
                EcmCsvName = f"Ecm-{grid_id}grid-{self.schedule.time + self.loaded_max_step}step.csv"
                path_to_save = os.path.join(self.new_simulation_folder, "Ecm", EcmCsvName)
                new_ecm_df.to_csv(path_to_save)

                df_time_grids_got_populated[f"Time when grid {grid_id} was first populated"] = [self.time_grid_got_populated[grid_id-1]]
                df_time_grids_got_populated_csv_name = f"Cells-are-present-grid-{grid_id}-{self.schedule.time + self.loaded_max_step}step.csv"
            path_to_save = os.path.join(self.new_simulation_folder, "Time when grids were populated", df_time_grids_got_populated_csv_name)
            df_time_grids_got_populated.to_csv(path_to_save)

            # Saves vasculature data
            # {key: list of clusters} -> {timestep: [(number of Mcells, number of Ecells), ..., (..., ...)]}
            vasculature_json = json.dumps(self.vasculature)
            
            vasculature_json_name = f"Vasculature-{self.schedule.time + self.loaded_max_step}step.json"
            path_to_save = os.path.join(self.new_simulation_folder, "Vasculature", vasculature_json_name)
            
            with open(path_to_save, 'w') as f:
                f.write(vasculature_json)
                
            # Saves cancer cells data as a backup in case the simulation fails
            # _, current_model_data = mesa.batchrunner._collect_data(self, self.data_collection_period-1)
            # df_current_model_data = pd.DataFrame(current_model_data)
            # df_current_model_data["Step"] = self.data_collection_period
            # for step in range(self.data_collection_period * 2, self.schedule.time, self.data_collection_period):
            #     _, step_model_data = mesa.batchrunner._collect_data(self, step-1)
            #     df_step_model_data = pd.DataFrame(step_model_data)
            #     df_step_model_data["Step"] = step + self.loaded_max_step
            #     df_current_model_data = pd.concat([df_current_model_data, df_step_model_data])
            # path_to_save = os.path.join(self.new_simulation_folder, f'CellsData.csv')
            # df_current_model_data.to_csv(path_to_save)



    def proliferate(self, cell_type):
        """"
        Duplicates every cell of cell_type phenotype in every site of the model

        Input: none
        Returns: none
        """
        for agent in self.schedule.agents:
            if agent.agent_type == "cell":
                x, y = agent.pos
                amount_of_cells = len([cell for cell in agent.grid.get_cell_list_contents([(x, y)]) if cell.agent_type == "cell"])
                if carrying_capacity > amount_of_cells and agent.phenotype == cell_type:
                    # print("Created new cell!!")
                    new_cell = CancerCell(self.current_agent_id, self, agent.grid, agent.grid_id, agent.phenotype, agent.ecm, agent.mmp2)
                    self.current_agent_id += 1
                    self.schedule.add(new_cell)
                    agent.grid.place_agent(new_cell, (x,y))
                    self.cancer_cells_counter[agent.grid_id - 1] += 1
        


    def load_previous_simulation(self, path_to_simulation):
        """
        Loads the last step of a previously computed simulation as the initial condition of this model

        Input: The simulation's path to be loaded
        Returns: none
        """
        #load mmp2 and ecm 
        mmp2_files_path = os.path.join(path_to_simulation, "Mmp2")
        ecm_files_path  = os.path.join(path_to_simulation, "Ecm")
        for grid_number in range(self.grids_number):
            mmp2_files = [file for file in os.listdir(mmp2_files_path) if file.startswith(f"Mmp2-{grid_number+1}grid")]
            ecm_files  = [file for file in os.listdir(ecm_files_path) if file.startswith(f"Ecm-{grid_number+1}grid")]
            mmp2_files.sort(key = lambda file_name: int(file_name.split('step')[0][11:]))
            ecm_files.sort(key  = lambda file_name: int(file_name.split('step')[0][10:]))
            last_state_of_mmp2_filepath = os.path.join(mmp2_files_path,mmp2_files[-1])
            last_state_of_ecm_filepath  = os.path.join(ecm_files_path,ecm_files[-1])
            print(f"Loading MMP2 state for grid id {grid_number + 1} in {last_state_of_mmp2_filepath}.")
            print(f"Loading ECM state for grid id {grid_number + 1} in {last_state_of_ecm_filepath}.")
            self.ecm[grid_number][0,:,:]  = pd.read_csv(last_state_of_ecm_filepath, index_col=0).to_numpy(dtype=float)
            self.mmp2[grid_number][0,:,:] = pd.read_csv(last_state_of_mmp2_filepath, index_col=0).to_numpy(dtype=float)

        path = os.path.join(path_to_simulation, "CellsData.csv")
        previous_sim_df = pd.read_csv(path, converters={"Position": ast.literal_eval})
        last_step = previous_sim_df["Step"].max()
        self.loaded_max_step = last_step
        previous_sim_df = previous_sim_df[previous_sim_df["Step"] == last_step]
        last_step_cells = previous_sim_df[previous_sim_df["Agent Type"] == "cell"]
        last_step_vessels = previous_sim_df[previous_sim_df["Agent Type"] == "vessel"]
        self.number_of_initial_cells = 0
        for index, row in last_step_cells.iterrows():
            current_grid_number = int(row["Grid"]) - 1
            ccell = CancerCell(self.current_agent_id, self, self.grids[current_grid_number], self.grid_ids[current_grid_number], row["Phenotype"], self.ecm[current_grid_number], self.mmp2[current_grid_number])
            self.current_agent_id += 1
            self.schedule.add(ccell)
            self.grids[current_grid_number].place_agent(ccell, row["Position"])
            self.cancer_cells_counter[current_grid_number] += 1
            self.number_of_initial_cells += 1
        for index, row in last_step_vessels.iterrows():
            current_grid_number = int(row["Grid"]) - 1
            ruptured_state = bool(row["Ruptured"])
            vessel = Vessel(self.current_agent_id, self, ruptured_state, self.grids[current_grid_number], self.grid_ids[current_grid_number])
            self.current_agent_id += 1
            self.schedule.add(vessel)
            self.grids[current_grid_number].place_agent(vessel, row["Position"])
            self.grid_vessels_positions[current_grid_number] += [row["Position"]]

        #load vasculature
        vasculature_path = os.path.join(path_to_simulation, "Vasculature")
        vasculature_files = os.listdir(vasculature_path)
        vasculature_files.sort(key = lambda file_name: int(file_name.split('step')[0][12:]))
        last_state_of_vasculature_filepath = os.path.join(vasculature_path,vasculature_files[-1])
        with open(last_state_of_vasculature_filepath, 'r') as f:
            last_state_of_vasculature = json.load(f)
        # Change keys to int
        last_state_of_vasculature = {int(k): v for k, v in last_state_of_vasculature.items()}
        self.vasculature = last_state_of_vasculature

        #calculate state of doubling counters
        self.doubling_time_counter_E = doubling_time_E - (last_step % doubling_time_E)
        self.doubling_time_counter_M = doubling_time_M - (last_step % doubling_time_M)

        #load time_grid_got_populated
        time_grid_got_populated_path = os.path.join(path_to_simulation, "Time when grids were populated")
        time_grid_got_populated_files = os.listdir(time_grid_got_populated_path)
        time_grid_got_populated_files.sort(key = lambda file_name: int(file_name.split('step')[0][25:]))
        time_grid_got_populated_filepath = os.path.join(time_grid_got_populated_path,time_grid_got_populated_files[-1])
        df_time_grid_got_populated = pd.read_csv(time_grid_got_populated_filepath, index_col=0)
        self.time_grid_got_populated = df_time_grid_got_populated.loc[0, :].values.flatten().tolist()


    def _initialize_grids(self):
        """
        Places the initial cancer cell and vessel in the initial grid in a circle

        Input: none
        Returns: none
        """
        mesenchymal_number = round(self.number_of_initial_cells * mesenchymal_proportion)
        possible_places = find_quasi_circle(n_center_points_for_tumor, self.width, self.height)[1]
        # Place all the agents in the quasi-circle area in the center of the grid
        for i in range(self.number_of_initial_cells):
            if mesenchymal_number > 0:
                cell_type = "mesenchymal"
                mesenchymal_number -= 1
            elif mesenchymal_number == 0:
                cell_type = "epithelial"

            a = CancerCell(self.current_agent_id, self, self.grids[0], self.grid_ids[0], cell_type, self.ecm[0], self.mmp2[0])
            self.current_agent_id += 1
            j = self.random.randrange(len(possible_places))
            x = int(possible_places[j][0])
            y = int(possible_places[j][1])

            self.schedule.add(a)
            self.grids[0].place_agent(a, (x, y))
            self.cancer_cells_counter[0] += 1

            # Remove the point after it has an amount of cells equal to the carrying capacity
            possible_places[j][2] += 1
            if possible_places[j][2] == carrying_capacity:
                possible_places.pop(j)


        # Create agents at second grid. Useful for debugging.
        amount_of_second_grid_cancer_cells = 0
        for i in range(amount_of_second_grid_cancer_cells):
            a = CancerCell(self.current_agent_id, self, self.grids[1], self.grid_ids[1], "mesenchymal", self.ecm[1], self.mmp2[1])
            self.current_agent_id += 1
            self.schedule.add(a)
        
            # Add the agent to a random grid cell
            x = self.random.randrange(3,7)
            y = self.random.randrange(3,7)
            self.grids[1].place_agent(a, (x, y))
            self.cancer_cells_counter[1] += 1

        # Create vessels
        num_normal_vessels = normal_vessels_primary
        num_ruptured_vessels = ruptured_vessels_primary

        # creates grid with 1 where vessels must not be placed
        not_possible_array = find_quasi_circle(n_center_points_for_Vessels, self.width, self.height)[0]
        not_possible_array[:2,:] = 1
        not_possible_array[-2:,:] = 1
        not_possible_array[:,:2] = 1
        not_possible_array[:,-2:] = 1
        possible_places = np.where(not_possible_array == 0)
        pos_coords = [list(tup) for tup in zip(possible_places[0], possible_places[1])]

        for i in range(len(self.grids)):
            if i == 0: # primary grid
                temp = num_ruptured_vessels
                while temp > 0:
                    j = num_ruptured_vessels - temp
                    coord_to_place = [self.random.randrange(self.width), self.random.randrange(self.height)]
                    if coord_to_place in pos_coords:
                        a = Vessel(self.current_agent_id, self, True, self.grids[0], self.grid_ids[0])
                        self.current_agent_id += 1
                        self.schedule.add(a)
                        self.grids[0].place_agent(a, (int(coord_to_place[0]), int(coord_to_place[1])))
                        self.grid_vessels_positions[i] += [(int(coord_to_place[0]), int(coord_to_place[1]))]
                        not_possible_array[coord_to_place[0], coord_to_place[1]] = 1
                        pos_coords.remove(coord_to_place)
                        temp -= 1

                temp = num_normal_vessels
                while temp > 0:
                    j = num_normal_vessels - temp
                    coord_to_place = [self.random.randrange(self.width), self.random.randrange(self.height)]
                    if coord_to_place in pos_coords:
                        a = Vessel(self.current_agent_id, self, False, self.grids[0], self.grid_ids[0])
                        self.current_agent_id += 1
                        self.schedule.add(a)
                        self.grids[0].place_agent(a, (int(coord_to_place[0]), int(coord_to_place[1])))
                        self.grid_vessels_positions[i] += [(int(coord_to_place[0]), int(coord_to_place[1]))]

                        not_possible_array[coord_to_place[0], coord_to_place[1]] = 1
                        pos_coords.remove(coord_to_place)
                        temp -= 1
            elif i > 0: # secondary grid and beyond
                    for m in range(secondary_sites_vessels[i-1]):
                        a = Vessel(self.current_agent_id, self, False, self.grids[i], self.grid_ids[i])
                        self.current_agent_id += 1
                        self.schedule.add(a)
                        x = self.random.randrange(self.width)
                        y = self.random.randrange(self.height)
                        self.grids[i].place_agent(a, (x,y))
                        self.grid_vessels_positions[i] += [(x,y)]
                
    def calculate_environment(self, mmp2, ecm):
        global th
        for i in range(len(mmp2)):
            for cell in self.grids[i].coord_iter():
                cell_contents, (x, y) = cell
                self.mesenchymal_count[i][x,y] = 0
                self.epithelial_count[i][x,y] = 0
                for cancer_cell in cell_contents:
                    if isinstance(cancer_cell, CancerCell):
                        if cancer_cell.phenotype == "mesenchymal":
                            self.mesenchymal_count[i][x,y] += 1
                        elif cancer_cell.phenotype == "epithelial":
                            self.epithelial_count[i][x,y] += 1
                        else:
                            raise Exception("Unknown phenotype")
                on_left_border = self.grids[i].out_of_bounds((x-1,y))
                on_right_border = self.grids[i].out_of_bounds((x+1,y))
                on_top_border = self.grids[i].out_of_bounds((x,y-1))
                on_bottom_border = self.grids[i].out_of_bounds((x,y+1))
                mmp2[i][1,x,y]=dmmp*tha/xha**2*((mmp2[i][0,x+1,y] if not on_right_border else mmp2[i][0,x-1,y])\
                        +(mmp2[i][0,x-1,y] if not on_left_border else mmp2[i][0,x+1,y])\
                        +(mmp2[i][0,x,y+1] if not on_bottom_border else mmp2[i][0,x,y-1])\
                        +(mmp2[i][0,x,y-1] if not on_top_border else mmp2[i][0,x,y+1])\
                        )\
                        +mmp2[i][0,x,y]*(1-4*dmmp*tha/xha**2-th*Lambda)+tha*theta*self.mesenchymal_count[i][x,y]
                ecm[i][1,x,y] = ecm[i][0,x,y]*(1-tha*(gamma1*self.mesenchymal_count[i][x,y]+gamma2*mmp2[i][1,x,y]))
                if ecm[i][1,x,y] < 0:
                    warnings.warn(f"<0 ecm in [i][1,{x},{y}] is {ecm[i][1,x,y]}")
                if ecm[i][1,x,y] > 1:
                    warnings.warn(f">1 ecm in [i][1,{x},{y}] is {ecm[i][1,x,y]}")
                    print("ECM is greater than 1! Your MMP2 diffusion rate is probably too high")
            mmp2[i][0,:,:] = mmp2[i][1,:,:]
            ecm[i][0,:,:] = ecm[i][1,:,:]

    def disaggregate_clusters(self, time):
        """
        Dissagregates cells from clusters into single-cell clusters, according to
        the dissagregation probability

        Input:
            time: given time for which the dissagreggation will occur
        Returns: 
            None
        """
        big_clusters = [cluster for cluster in self.vasculature[time] if sum(cluster) > 1]
        new_vasculature = [cluster for cluster in self.vasculature[time] if sum(cluster) == 1]
        for cluster in big_clusters:
            new_mesenchymal, new_epithelial = cluster
            for ccell_type, ccells_amount in enumerate(cluster):
                for i in range(ccells_amount):
                    if self.random.random() > dissagreggation_prob:
                        if ccell_type == 0:
                            new_vasculature += [(1, 0)]
                            new_mesenchymal -= 1
                        if ccell_type == 1:
                            new_vasculature += [(0, 1)]
                            new_epithelial -= 1
            if new_mesenchymal + new_epithelial > 0:
                new_vasculature += [(new_mesenchymal,new_epithelial)]
        self.vasculature[time] = new_vasculature


from pynput import keyboard
import os
import metaspread.datagenerator as datagenerator
import metaspread.graphgenerator as graphgenerator
import metaspread.videogenerator as videogenerator
import metaspread.simrunner as simrunner
import time

selected_option_index = 0
selected_option = ""
options_list =  ["New simulation", "Load simulation", "Postprocessing", "Exit"]
banner_message = "Welcome to MetaSpread: a Cancer Simulation Program!"

def main_menu():
    with keyboard.Listener(on_press = on_press) as listener:
        os.system('cls')
        print_menu()
        listener.join()
        if selected_option == "New simulation":
            simulation_menu()
        elif selected_option == "Load simulation":
            load_simulation_menu()
        elif selected_option == "Postprocessing":
            postprocessing_menu()
        elif selected_option == "Exit":
            os._exit(0)


def simulation_menu():
    print("==================================================================")
    print("New simulation")
    print("==================================================================")
    total_steps = get_int_input("Select the total timesteps the simulation will take: ") #convert this to time and show, asking if it;s ok instead of the other comments
    interval_steps = get_int_input("Select the size of the intervals for which the information will be collected: ")
    simrunner.run_simulation(total_steps, interval_steps)


def load_simulation_menu():
    global selected_option, selected_option_index, banner_message, options_list
    directory_path = "Simulations"
    folder_names = ["Exit"] + get_folder_names(directory_path)
    options_list = folder_names
    selected_option_index = 0
    banner_message = "Load simulation: select simulation"
    with keyboard.Listener(on_press = on_press) as listener:
        os.system('cls')
        print_menu()
        listener.join()
        if selected_option == "Exit":
            os._exit(0)
    total_steps = get_int_input("Select additional number of steps until the loaded simulation will continue: ")
    interval_steps = get_int_input("Select the size of the intervals for which the information will be collected: ")
    print("Loading simulation...")
    simrunner.run_simulation(total_steps, interval_steps, loaded_simulation_path=os.path.join(os.getcwd(),directory_path, selected_option))

def postprocessing_menu():
    global selected_option, selected_option_index, banner_message, options_list
    directory_path = "Simulations"
    folder_names = ["Exit"] + get_folder_names(directory_path)
    options_list = folder_names
    selected_option_index = 0
    banner_message = "Postprocessing: select simulation"
    with keyboard.Listener(on_press = on_press) as listener:
        os.system('cls')
        print_menu()
        listener.join()
        if selected_option == "Exit":
            os._exit(0)
    options_list = ["Exit", "Run all", "Begin data analysis", "Begin graphical analysis", "Generate videos"]
    selected_simulation = selected_option
    selected_option_index = 0
    banner_message = f"Postprocessing simulation at {selected_simulation}"
    with keyboard.Listener(on_press = on_press) as listener:
        os.system('cls')
        print_menu()
        listener.join()
        if selected_option == "Exit":
            os._exit(0)
        if selected_option == "Run all":
            amount_of_pictures = get_int_input(f"Select the amount of pictures that will be produced for the graphical analysis: ")
            frame_rate = get_int_input("Select the framerate for the video.\nA framerate of 10 is sugggested \nfor every 50 pictures:")
            datagenerator.generate_data(selected_simulation)
            time.sleep(3)
            graphgenerator.generate_graphs(selected_simulation, amount_of_pictures)
            time.sleep(3)
            videogenerator.generate_videos(selected_simulation, frame_rate)
            time.sleep(3)
        if selected_option == "Begin data analysis":
            datagenerator.generate_data(selected_simulation)
            time.sleep(3)
        if selected_option == "Begin graphical analysis":
            graphical_analysis_menu(selected_simulation)
        if selected_option == "Generate videos":
            frame_rate = get_int_input("Select the framerate for the video.\nA framerate of 10 is sugggested \nfor every 50 pictures:")
            videogenerator.generate_videos(selected_simulation, frame_rate)
            time.sleep(3)



def graphical_analysis_menu(selected_simulation):
    global selected_option, selected_option_index, banner_message, options_list
    directory_path = "Simulations"
    options_list = ["Generate 5 picture summary", "Custom amount of pictures", "Generate All"]
    selected_option_index = 0
    banner_message = "Graphical analysis: select which frames are going to be plotted"
    with keyboard.Listener(on_press = on_press) as listener:
        os.system('cls')
        print_menu()
        listener.join()
        if selected_option == "Exit":
            os._exit(0)
        if selected_option == "Generate 5 picture summary":
            graphgenerator.generate_graphs(selected_simulation, 5)
            time.sleep(3)
        if selected_option == "Custom amount of pictures":
            amount_of_pictures = get_int_input(f"Select the amount of pictures that will be produced: ")
            graphgenerator.generate_graphs(selected_simulation, amount_of_pictures)
            time.sleep(3)
        if selected_option == "Generate All":
            graphgenerator.generate_graphs(selected_simulation)
            time.sleep(3)


def print_menu():
    if banner_message != "":
        print("==================================================================")
        print(banner_message)
        print("==================================================================")
    for i, option in enumerate(options_list):
        if i == selected_option_index:
            print("➤", option)
        else:
            print("  ", option)

def on_press(key):
    global selected_option_index
    if key == keyboard.Key.down:
        if selected_option_index < len(options_list) - 1:
            selected_option_index += 1
            os.system('cls')
            print_menu()
    elif key == keyboard.Key.up:
        if selected_option_index > 0:
            selected_option_index -= 1
            os.system('cls')
            print_menu()
    elif key == keyboard.Key.enter:
        global selected_option
        selected_option = options_list[selected_option_index]
        os.system('cls')
        return False
    elif key == keyboard.Key.esc:
        os._exit(0)

def get_folder_names(directory_path):
    folder_names = []
    for folder_name in os.listdir(directory_path):
        if os.path.isdir(os.path.join(directory_path, folder_name)):
            folder_names.append(folder_name)
    print(folder_names)
    folder_names.sort(key=lambda x: os.path.getmtime(f"{directory_path}\{x}"), reverse=True)
    return folder_names

def get_int_input(text):
    result = ''
    while not isinstance(result, int) or result < 0:
        while result == '':
            os.system('cls')
            result = input(text)
            try:
                result = int(result)
            except:
                result = ''
    return result

if __name__ == "__main__":
    main_menu()


import cv2
import os

# To run this code you must be in the parent folder of agent-based-cancer

def group_images_by_grid(images_folder_path):
    """
    Groups all images (.png) on the images_folder_path by grid and returns a list of lists,
    where each list contains the path to the corresponding grid.
    """
    # Initialize dictionary to store paths of images by grid
    images_by_grid = {}

    # Iterate through all images in the folder
    for image_filename in os.listdir(images_folder_path):
        #print(f'image_filename: {image_filename}')
        # Skip non-png files
        if not image_filename.endswith(".png"):
            continue
        
        # Extract grid and timestep from image filename
        grid = image_filename.split("-")[1]
        timestep = image_filename.split("-")[2].split(" ")[0]
        
        # Create list for the grid if it doesn't exist yet
        if grid not in images_by_grid:
            images_by_grid[grid] = []
        
        # Add path to image to the list for the corresponding grid
        image_path = os.path.join(images_folder_path, image_filename)
        images_by_grid[grid].append((timestep, image_path))
    #print(f'IMAGES_BY_GRID: {images_by_grid}')
    # Sort images in each grid list by timestep
    for grid in images_by_grid:
        images_by_grid[grid].sort(key=lambda x: int(x[0].split("step")[-1]))

    # Convert dictionary to list of lists
    images_by_grid_list = [images_by_grid[grid] for grid in sorted(images_by_grid)]

    return images_by_grid_list


def create_video_from_images(images_list, video_path, frameRate):
    """
    creates a video
    showing the dynamics of the images over time.
    """
    # Initialize video writer
    img = cv2.imread(images_list[0])
    height, width, channels = img.shape
    fourcc = cv2.VideoWriter_fourcc(*'mp4v')
    fps = frameRate
    out = cv2.VideoWriter(video_path, fourcc, fps, (width, height), isColor=True)

    # Add each image to the video
    for image_path in images_list:
        image = cv2.imread(image_path)
        out.write(image)

    # Release video writer
    out.release()
    print(f"\tGenerated video: {video_path}")



def create_combined_video(images_by_grid_list, video_path, frameRate):
    """
    Given a list of lists of (timestep, image_path) tuples, creates a video
    showing the dynamics of the images over time, with one row for each grid.
    """
    # Initialize video writer
    img = cv2.imread(images_by_grid_list[0][0][1])
    height, width, channels = img.shape
    fourcc = cv2.VideoWriter_fourcc(*'mp4v')
    fps = frameRate
    out = cv2.VideoWriter(video_path, fourcc, fps, (width, height), isColor=True)

    # Add each image to the combined video
    for i in range(len(images_by_grid_list[0])):
        # Create a single frame for this timestep
        frame = None
        for j, grid_images in enumerate(images_by_grid_list):
            _, image_path = grid_images[i]
            image = cv2.imread(image_path)
            image = cv2.resize(image, (width // 2, height // 2))
            if frame is None:
                frame = image
            else:
                frame = cv2.hconcat([frame, image])
        # Add frame to combined video
        for i in range(3):
            out.write(frame)

    # Release video writer
    out.release()
    print(f"\tGenerated video: {video_path}")

def generate_videos(nameOfTheSimulation, frameRate):
    
    # Path where the folder of images are located
    imagesFolder = "Graphical analysis"
    simulations_dir = "Simulations"
    ImagesFolderPath = os.path.join(simulations_dir, nameOfTheSimulation, imagesFolder)
    print(f'Analyzing images at: {ImagesFolderPath}')

    #Path to save the videos
    VideosFolderPath = os.path.join(simulations_dir, nameOfTheSimulation, "Videos")
    print(f'Creating folder to save videos at: {VideosFolderPath}\n')

    print(f'\tFrame rate is: {frameRate}')

    # Path of the images (.png)
    EcmImagesPath = os.path.join(ImagesFolderPath, "Ecm dynamics")
    Mmp2ImagesPath = os.path.join(ImagesFolderPath, "Mmp2 dynamics")
    CellsImagesPath = os.path.join(ImagesFolderPath, "Tumor dynamics")
    CellsNumberImagesPath = os.path.join(ImagesFolderPath, "Cells growth")
    VasculatureImagesPath = os.path.join(ImagesFolderPath, "Vasculature dynamics")

    # Create folder for all the videos
    if not os.path.exists(VideosFolderPath):
        os.makedirs(VideosFolderPath)
    # If there the Graphical analysis is already done, tell the user
    else:
        print(f"(Videos folder already exists)! \n\tVideos that already exist will not be overwritten)")

    # Create list of lists. Each list will contain all the path to the Ecm of a given grid
    EcmImagesByGrid = (group_images_by_grid(EcmImagesPath))
    Mmp2ImagesByGrid = (group_images_by_grid(Mmp2ImagesPath))
    CellsImagesByGrid = (group_images_by_grid(CellsImagesPath))
    CellsNumberImagesByGrid = (group_images_by_grid(CellsNumberImagesPath))

    # Create video for Vasculature dynamics
    imagesFileNames = os.listdir(VasculatureImagesPath)
    imagesFileNames.sort()
    images_vasculature_cells = imagesFileNames[:len(imagesFileNames)//2]
    images_vasculature_clusters = imagesFileNames[len(imagesFileNames)//2:]
    images_vasculature_cells.sort(key = lambda x: int(x.split('.')[0][22:]))
    images_vasculature_clusters.sort(key = lambda x: int(x.split('.')[0][25:]))
    images_vasculature_cells    = [os.path.join(VasculatureImagesPath, file_name) for file_name in images_vasculature_cells]
    images_vasculature_clusters = [os.path.join(VasculatureImagesPath, file_name) for file_name in images_vasculature_clusters]
    output_file = os.path.join(VideosFolderPath, f"Vasculature - cells.mp4")
    create_video_from_images(images_vasculature_cells, output_file, frameRate)
    output_file = os.path.join(VideosFolderPath, f"Vasculature - clusters.mp4")
    create_video_from_images(images_vasculature_clusters, output_file, frameRate)

    # Create videos for the Ecm of each grid
    for i, images_list in enumerate(EcmImagesByGrid):
        steps = tuple(image[0] for image in images_list)
        imagesPath = tuple(image[1] for image in images_list)
        output_file = os.path.join(VideosFolderPath, f"Ecm dynamics - Grid{i+1}.mp4")
        create_video_from_images(imagesPath, output_file, frameRate)

    # Create videos for the Mmp2 of each grid
    for i, images_list in enumerate(Mmp2ImagesByGrid):
        steps = tuple(image[0] for image in images_list)
        imagesPath = tuple(image[1] for image in images_list)
        output_file = os.path.join(VideosFolderPath, f"Mmp2 dynamics - Grid{i+1}.mp4")
        create_video_from_images(imagesPath, output_file, frameRate)

    # Create videos for the Tumor of each grid
    for i, images_list in enumerate(CellsImagesByGrid):
        steps = tuple(image[0] for image in images_list)
        imagesPath = tuple(image[1] for image in images_list)
        output_file = os.path.join(VideosFolderPath, f"Tumor dynamics - Grid{i+1}.mp4")
        create_video_from_images(imagesPath, output_file, frameRate)

    # Create video for the cell number growth
    for i, images_list in enumerate(CellsNumberImagesByGrid):
        steps = tuple(image[0] for image in images_list)
        imagesPath = tuple(image[1] for image in images_list)
        output_file = os.path.join(VideosFolderPath, f"Cells Number - Grid{i+1}.mp4")
        create_video_from_images(imagesPath, output_file, frameRate)

    print(f'\nFinished!')





import metaspread.configs
import pandas as pd
import shutil
import mesa
import ast
import os
from pathlib import Path

# To run this code you must be in the parent folder of the program

def save_configs(simulations_dir, new_simulation_folder, config_var_names, max_steps, data_collection_period):
    # Saves the simulation configuration
    print(f"\t Saving all the simulations parameters at: {os.path.join(simulations_dir, new_simulation_folder, 'configs.csv')}")
    values = [getattr(metaspread.configs, i) for i in config_var_names]
    names = config_var_names

    #add configurations that are not in the global variables
    names += ['max_steps', 'data_collection_period']
    values += [max_steps, data_collection_period]
    df_vars = pd.DataFrame({"Names": names, "Values": values})
    df_vars = df_vars.set_index("Names")
    path = os.path.join(simulations_dir, new_simulation_folder, 'configs.csv')
    df_vars.to_csv(path)


def run_simulation(max_steps, data_collection_period, save_path=Path("."), loaded_simulation_path=""):

    # load configs file from a previous simulation or loads the general configs file
    print(loaded_simulation_path)
    loaded_simulation_path= loaded_simulation_path.strip('\"')
    if loaded_simulation_path != "":
        configs_path = os.path.join(loaded_simulation_path, "configs.csv")
        config_var_names = metaspread.configs.load_simulation_configs_for_reloaded_simulation(configs_path)
    else:
        configs_path = "simulations_configs.csv"
        config_var_names = metaspread.configs.init_simulation_configs(configs_path)
    
    # Parameters for this simulation
    number_of_initial_cells = metaspread.configs.number_of_initial_cells # Number of cancer cells
    gridsize     = metaspread.configs.gridsize
    grids_number = metaspread.configs.grids_number
    width        = gridsize
    height       = gridsize

    # Name of the directories
    simulations_dir = save_path / "Simulations"
    os.makedirs(simulations_dir, exist_ok=True)
    if loaded_simulation_path != "":
        cells_path = os.path.join(loaded_simulation_path, "CellsData.csv")
        df = pd.read_csv(cells_path)
        loaded_max_step = max(df["Step"])
        new_simulation_folder = os.path.normpath(loaded_simulation_path)
        new_simulation_folder = os.path.basename(new_simulation_folder)
        new_simulation_path = os.path.join(simulations_dir, new_simulation_folder)
    else:
        df = pd.DataFrame()
        loaded_max_step = 0
        new_simulation_folder = f"Sim-max_steps-{max_steps}-collection_period-{data_collection_period}-cells-{number_of_initial_cells}-grids_number-{grids_number}"

        # Creates the path for the new simulation
        new_simulation_path = os.path.join(simulations_dir, new_simulation_folder)
        pathMmp2 = os.path.join(new_simulation_path, "Mmp2")
        pathEcm = os.path.join(new_simulation_path, "Ecm")
        pathVasculature = os.path.join(new_simulation_path, "Vasculature")
        pathTimeOfPopulation = os.path.join(new_simulation_path, "Time when grids were populated")

        # Create folder for all cells analysis, for Mmp2 matrices and Ecm matrices
        if not os.path.exists(new_simulation_path):
            print(f'\t Folder for this simulation: {new_simulation_path}')
            print(f'\t Saving agents data at: {new_simulation_path}')
            print(f'\t Saving Mmp2 data at: {pathMmp2}')
            print(f'\t Saving Ecm data at: {pathEcm}')
            print(f'\t Saving Vasculature data at: {pathVasculature}')

            os.makedirs(new_simulation_path)
            os.makedirs(pathMmp2)
            os.makedirs(pathEcm)
            os.makedirs(pathVasculature)
            os.makedirs(pathTimeOfPopulation)
        # If there is already a simulation you skip it
        else:
            return print("This simulation already exists!")

    # Run the simulation and saves the data
    save_configs(simulations_dir, new_simulation_folder, config_var_names, max_steps, data_collection_period)
    model = metaspread.CancerModel(
        number_of_initial_cells,
        width,
        height,
        grids_number,
        max_steps,
        data_collection_period,
        new_simulation_path,
        loaded_simulation_path)
    for i in range(max_steps):
        model.step()
    print(f'Finished the simulation at time step {model.schedule.time}!')
    return model


from metaspread.configs import init_simulation_configs, check_if_configs_are_present
check_if_configs_are_present()
init_simulation_configs("simulations_configs.csv")
from metaspread.cancercell import CancerCell
from metaspread.cancermodel import CancerModel
from metaspread.vessel import Vessel


import mesa
from metaspread.vessel import Vessel
from metaspread.configs import *

class CancerCell(mesa.Agent):

    def __init__(self, unique_id, model, grid, grid_id, phenotype, ecm, mmp2):
        super().__init__(unique_id, model)
        self.grid = grid
        self.grid_id = grid_id
        self.phenotype = phenotype
        if self.phenotype == "mesenchymal":
            self.diff_coeff = dM
            self.phi = phiM
        else:
            self.diff_coeff = dE
            self.phi = phiE
        self.ecm = ecm
        self.mmp2 = mmp2
        self.agent_type = "cell"
        self.ruptured = False #need to be able do use data collector on agents
        
    def step(self): #what will the agent do every time a step is made
        self.move()

    def move(self):
        #fixed probabilities can be given to fix the movement of the cells towards a certain direction
        #if no fixed probabilities are given, the probabilities are calculated based on the ECM and MMP-2 concentration
        #(the fixed probabilities are only used for testing)
        fixed_p_left    = self.model.fixed_p_left
        fixed_p_right   = self.model.fixed_p_right
        fixed_p_top     = self.model.fixed_p_top
        fixed_p_bottom  = self.model.fixed_p_bottom
        time = self.model.schedule.time
        possible_steps = self.grid.get_neighborhood(
            self.pos,
            moore=False,
            include_center=True)
        x, y = self.pos
        on_left_border    = self.grid.out_of_bounds((x-1,y))
        on_right_border   = self.grid.out_of_bounds((x+1,y))
        on_top_border     = self.grid.out_of_bounds((x,y+1))
        on_bottom_border  = self.grid.out_of_bounds((x,y-1))
        p_left   = (fixed_p_left if fixed_p_left is not None
                        else 0 if on_left_border 
                        else (th/xh**2*(self.diff_coeff-self.phi/4*
                            (0  if on_right_border
                                else self.ecm[0,x+1,y]-self.ecm[0,x-1,y]))))
        p_right  = (fixed_p_right if fixed_p_right is not None
                        else 0 if on_right_border 
                        else (th/xh**2*(self.diff_coeff+self.phi/4*
                            (0  if on_left_border
                                else self.ecm[0,x+1,y]-self.ecm[0,x-1,y]))))
        p_top    = (fixed_p_top if fixed_p_top is not None
                        else 0 if on_top_border 
                        else (th/xh**2*(self.diff_coeff+self.phi/4*
                            (0  if on_bottom_border
                                else self.ecm[0,x,y+1]-self.ecm[0,x,y-1]))))
        p_bottom = (fixed_p_bottom if fixed_p_bottom is not None
                        else 0 if on_bottom_border 
                        else (th/xh**2*(self.diff_coeff-self.phi/4*
                            (0  if on_top_border
                                else self.ecm[0,x,y+1]-self.ecm[0,x,y-1]))))
        p_stay      = 1-(p_left+p_right+p_top+p_bottom)

        weights=[]
        for x2,y2 in possible_steps:
            if x2 < x:
                weights.append(p_left)
            elif x2>x:
                weights.append(p_right)
            elif y2<y:
                weights.append(p_bottom)
            elif y2>y:
                weights.append(p_top)
            else:
                weights.append(p_stay)
        #there is a chance that the probability should be calculated at the end as 1-sum(weights), have to check


        # new_position = (x,y+1)
        new_position = self.random.choices(possible_steps,weights,k=1)[0]
        is_vessel = False
        is_ruptured = False
        for agent in self.grid.get_cell_list_contents([(new_position)]):
            if isinstance(agent, Vessel):
                is_ruptured = agent.ruptured
                is_vessel=True
                break
        if is_vessel and self.grid_id == 1 and (is_ruptured or self.phenotype == "mesenchymal"): 
                x, y = new_position
                on_left_border    = self.grid.out_of_bounds((x-1,y))
                on_right_border   = self.grid.out_of_bounds((x+1,y))
                on_top_border     = self.grid.out_of_bounds((x,y+1))
                on_bottom_border  = self.grid.out_of_bounds((x,y-1))
                mesenchymal_ccells_to_travel  = [agent for agent in self.grid.get_cell_list_contents([(x,y)]) if agent.agent_type == 'cell' and agent.phenotype == "mesenchymal"]
                mesenchymal_ccells_to_travel += [] if on_left_border   else [agent for agent in self.grid.get_cell_list_contents([(x-1,y)]) if agent.agent_type == 'cell' and agent.phenotype == "mesenchymal"]
                mesenchymal_ccells_to_travel += [] if on_right_border  else [agent for agent in self.grid.get_cell_list_contents([(x+1,y)]) if agent.agent_type == 'cell' and agent.phenotype == "mesenchymal"]
                mesenchymal_ccells_to_travel += [] if on_top_border    else [agent for agent in self.grid.get_cell_list_contents([(x,y-1)]) if agent.agent_type == 'cell' and agent.phenotype == "mesenchymal"]
                mesenchymal_ccells_to_travel += [] if on_bottom_border else [agent for agent in self.grid.get_cell_list_contents([(x,y+1)]) if agent.agent_type == 'cell' and agent.phenotype == "mesenchymal"]
                epithelial_ccells_to_travel  = [agent for agent in self.grid.get_cell_list_contents([(x,y)]) if agent.agent_type == 'cell' and agent.phenotype == "epithelial"]
                epithelial_ccells_to_travel += [] if on_left_border   else [agent for agent in self.grid.get_cell_list_contents([(x-1,y)]) if agent.agent_type == 'cell' and agent.phenotype == "epithelial"]
                epithelial_ccells_to_travel += [] if on_right_border  else [agent for agent in self.grid.get_cell_list_contents([(x+1,y)]) if agent.agent_type == 'cell' and agent.phenotype == "epithelial"]
                epithelial_ccells_to_travel += [] if on_top_border    else [agent for agent in self.grid.get_cell_list_contents([(x,y-1)]) if agent.agent_type == 'cell' and agent.phenotype == "epithelial"]
                epithelial_ccells_to_travel += [] if on_bottom_border else [agent for agent in self.grid.get_cell_list_contents([(x,y+1)]) if agent.agent_type == 'cell' and agent.phenotype == "epithelial"]
        
                #if there are not clusters at that time in the vasculature dict, create a new key for that time
                #and add the tuple

                if self.model.vasculature.get(time + vasculature_time,False):
                    self.model.vasculature[time + vasculature_time] += [(len(mesenchymal_ccells_to_travel), len(epithelial_ccells_to_travel))]
                # if there are clusters, add the tuple to that key
                else:
                    self.model.vasculature[time + vasculature_time] = [(len(mesenchymal_ccells_to_travel), len(epithelial_ccells_to_travel))]
                for ccell in mesenchymal_ccells_to_travel + epithelial_ccells_to_travel:
                    ccell.grid.remove_agent(ccell)
                    ccell.model.schedule.remove(ccell)
        else:
            if carrying_capacity > len([cell for cell in self.grid.get_cell_list_contents([new_position]) if agent.agent_type == 'cell']):
                self.grid.move_agent(self, new_position)



import mesa

class Vessel(mesa.Agent):
    def __init__(self, unique_id, model, ruptured, grid, grid_id):
        super().__init__(unique_id, model)
        self.model = model
        self.ruptured = ruptured
        self.grid = grid
        self.grid_id = grid_id
        self.agent_type = "vessel"
        self.phenotype = False #needed to be able do use data collector on agents

    def step(self):
        pass


import os
import sys
import warnings
import metaspread.simrunner as simrunner
import metaspread.datagenerator as datagenerator
import metaspread.graphgenerator as graphgenerator
import metaspread.videogenerator as videogenerator

if __name__ == "__main__":
    # simple checks for misspellings in the arguments
    if len(sys.argv) >= 3 and sys.argv[1] == "postprocess":
        if sys.argv[2] == "graphic":
            raise Exception("\nError! No known command for postprocessing named \'graphic\'. Did you mean \'graphics\'?")
        if sys.argv[2] == "video":
            raise Exception("\nError! No known command for postprocessing named\'video\'. Did you mean \'videos\'?")
        elif sys.argv[2] != "data" and sys.argv[2] != "videos" and sys.argv[2] != "graphics" and sys.argv[2] != "all":
            error_string = f"\nError! No known command for postprocessing named \'{sys.argv[2]}\'. Use \'data\', \'videos\', \'graphics\' or \'all\'."
            raise Exception(error_string)
        
    # actual processing of the arguments
    if len(sys.argv) == 1:
        from metaspread.interactive import *
        main_menu()
    elif len(sys.argv) == 2:
        if sys.argv[1] == "interactive":
            from metaspread.interactive import *
            main_menu()
        else:
            raise Exception("Incorrent amount of or unrecognized arguments!")
    elif len(sys.argv) == 3:
        raise Exception("Incorrent amount of or unrecognized arguments!")
    elif len(sys.argv) == 4:
        if sys.argv[1] == "run":
            total_steps     = int(sys.argv[2])
            interval_steps  = int(sys.argv[3])
            simrunner.run_simulation(total_steps, interval_steps)
        elif sys.argv[1] == "postprocess" and sys.argv[2] == "data":
                simulation_folder = sys.argv[3]
                datagenerator.generate_data(simulation_folder)
        elif sys.argv[1] == "postprocess" and sys.argv[2] == "graphics":
            simulation_folder = sys.argv[3]
            amount_of_pictures = 0 #this will generate all pictures!
            warnings.warn("No amount of pictures given to graphics. Generating all pictures.")
            graphgenerator.generate_graphs(simulation_folder, amount_of_pictures)
        else:
            raise Exception("Incorrent amount of or unrecognized arguments!")
    elif len(sys.argv) == 5:
        if sys.argv[1] == "load":
            simulation_folder = os.path.join("Simulations",sys.argv[2])
            total_steps     = int(sys.argv[3])
            interval_steps  = int(sys.argv[4])
            simrunner.run_simulation(total_steps, interval_steps, simulation_folder)
        elif sys.argv[1] == "postprocess" and sys.argv[2] == "graphics":
            simulation_folder = sys.argv[3]
            amount_of_pictures = int(sys.argv[4])
            graphgenerator.generate_graphs(simulation_folder, amount_of_pictures)
        elif sys.argv[1] == "postprocess" and sys.argv[2] == "videos":
            simulation_folder = sys.argv[3]
            frame_rate = int(sys.argv[4])
            videogenerator.generate_videos(simulation_folder, frame_rate)
    elif len(sys.argv) == 6:
        if sys.argv[1] == "postprocess" and sys.argv[2] == "all":
                simulation_folder  = sys.argv[3]
                amount_of_pictures = int(sys.argv[4])
                frame_rate = int(sys.argv[5])
                datagenerator.generate_data(simulation_folder)
                graphgenerator.generate_graphs(simulation_folder, amount_of_pictures)
                videogenerator.generate_videos(simulation_folder, frame_rate)
        else:
            raise Exception("Incorrent amount or unrecognized arguments!")
    else:
        print(sys.argv)
        raise Exception("Incorrent amount of or unrecognized arguments!")


import ast
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
import pandas as pd
import numpy as np
import re
import os
import json
import metaspread.configs

# To run this code you must be in the parent folder of agent-based-cancer

def read_coords_for_plot(step, all_cells_dataframe, grid_id):
    df = all_cells_dataframe[["Step", "Position", "Phenotype", "Grid", "Agent Type", "Ruptured"]]

    # Selected step to plot
    df_selected_step_and_grid = df.loc[(df["Step"] == step) & (df["Grid"] == grid_id)]

    # Save the position data
    mPoints = list(df_selected_step_and_grid.loc[df_selected_step_and_grid["Phenotype"] == "mesenchymal"]["Position"])
    ePoints = list(df_selected_step_and_grid.loc[df_selected_step_and_grid["Phenotype"] == "epithelial"]["Position"])
    vPoints = list(df_selected_step_and_grid.loc[(df_selected_step_and_grid["Agent Type"] == "vessel") & (df_selected_step_and_grid["Ruptured"] == False)]["Position"])
    vRupturedPoints = list(df_selected_step_and_grid.loc[(df_selected_step_and_grid["Agent Type"] == "vessel") & (df_selected_step_and_grid["Ruptured"] == True)]["Position"])

    X_position_m, Y_position_m = [i[0] for i in mPoints], [i[1] for i in mPoints]
    X_position_e, Y_position_e = [i[0] for i in ePoints], [i[1] for i in ePoints]
    X_position_v, Y_position_v = [i[0] for i in vPoints], [i[1] for i in vPoints]
    X_position_vr, Y_position_vr = [i[0] for i in vRupturedPoints], [i[1] for i in vRupturedPoints]

    return [X_position_m, Y_position_m, X_position_e, Y_position_e, X_position_v, Y_position_v, X_position_vr, Y_position_vr]

def save_cancer(all_cells_dataframe, grid_id, step, real_time_at_step, tumor_data_path):
    coords_list = False
    path = os.path.join(tumor_data_path, f'Cells-grid{grid_id}-step{step} - Tumor size at {real_time_at_step/(3600*24):.2f} days.csv')
    if not os.path.isfile(path):
        coords_list = read_coords_for_plot(step, all_cells_dataframe, grid_id)
        Xm, Ym, Xe, Ye, Xv, Yv, Xvr, Yvr = coords_list[0], coords_list[1], coords_list[2], coords_list[3], coords_list[4], coords_list[5], coords_list[6], coords_list[7]

        # save the data
        df_export = pd.DataFrame([Xm, Ym, Xe, Ye, Xv, Yv, Xvr, Yvr]) 
        df_export.to_csv(path)

    #create histogram of positions
    path = os.path.join(tumor_data_path, f'Cells-grid{grid_id}-step{step} - Histogram at {real_time_at_step/(3600*24):.2f} days.csv')
    if not os.path.isfile(path):
        if coords_list:
            Xm, Ym, Xe, Ye, Xv, Yv, Xvr, Yvr = coords_list[0], coords_list[1], coords_list[2], coords_list[3], coords_list[4], coords_list[5], coords_list[6], coords_list[7]
        else:
            coords_list = read_coords_for_plot(step, all_cells_dataframe, grid_id)
            Xm, Ym, Xe, Ye, Xv, Yv, Xvr, Yvr = coords_list[0], coords_list[1], coords_list[2], coords_list[3], coords_list[4], coords_list[5], coords_list[6], coords_list[7]
        df_positions = pd.DataFrame({'Position': zip(Xm + Xe, Ym + Ye)})
        position_repetition_count = df_positions['Position'].value_counts()
        histogram = position_repetition_count.value_counts()
        histogram = pd.DataFrame({'Bins': histogram.index, 'Frequency': histogram.values})
        number_of_empty_positions = metaspread.configs.gridsize * metaspread.configs.gridsize - len(position_repetition_count)
        new_row = pd.DataFrame({'Bins': [0], 'Frequency': [number_of_empty_positions]})
        histogram = pd.concat([histogram, new_row])
        histogram.to_csv(path)
        # return
        #this will return the radius and diameter of the processed tumor
        return get_cluster_centroid_radius_and_diameter(df_positions, grid_id)
    path = os.path.join(tumor_data_path, "Tumor radius and diameter history in grid 1.csv")
    if not os.path.isfile(path):
        if coords_list:
            Xm, Ym, Xe, Ye, Xv, Yv, Xvr, Yvr = coords_list[0], coords_list[1], coords_list[2], coords_list[3], coords_list[4], coords_list[5], coords_list[6], coords_list[7]
        else:
            coords_list = read_coords_for_plot(step, all_cells_dataframe, grid_id)
            Xm, Ym, Xe, Ye, Xv, Yv, Xvr, Yvr = coords_list[0], coords_list[1], coords_list[2], coords_list[3], coords_list[4], coords_list[5], coords_list[6], coords_list[7]
        df_positions = pd.DataFrame({'Position': zip(Xm + Xe, Ym + Ye)})
        return get_cluster_centroid_radius_and_diameter(df_positions, grid_id)
    return ([np.nan, np.nan], np.nan, np.nan)
    
def save_growth_data(all_cells_dataframe, grid_id, cells_data_path, step_number, real_time_at_step, real_delta_time, df_csv_last_step):
    path_to_save = os.path.join(cells_data_path, f'CellsGrowth-grid{grid_id}-step{step_number} - {real_time_at_step/(3600*24):.2f} days.csv')
    if os.path.isfile(path_to_save):
        return pd.DataFrame()
    df = all_cells_dataframe.loc[(all_cells_dataframe["Grid"] == grid_id) & (all_cells_dataframe["Step"] == step_number)]
    amount_of_mesenchymal = len(df[df["Phenotype"] == "mesenchymal"])
    amount_of_epithelial = len(df[df["Phenotype"] == "epithelial"])
    df = pd.DataFrame({"Number of Epithelial Cells": [amount_of_epithelial], "Number of Mesenchymal Cells": [amount_of_mesenchymal], "Steps": [step_number], "Days": [real_delta_time*step_number/(3600*24)]})
    if not df_csv_last_step.empty:
        df_csv = pd.concat([df_csv_last_step, df], ignore_index=True)
    else:
        df_csv = df
    df_csv.to_csv(path_to_save)
    return df_csv

def get_vasculature_state_at_step(path_to_save, vasculature_json_path, step):
    # Reads the dict in the json file
    with open(vasculature_json_path, 'r') as f:
        vasculature_dict = json.load(f)

    # Change keys to int
    vasculature_dict = {int(k): v for k, v in vasculature_dict.items()}

    # Prepare the data for the bar chart
    mesenchymal_count = 0
    epithelial_count = 0
    multicellular_cluster_count = 0
    total_cluster_count = 0
    time_steps = sorted(vasculature_dict.keys())
    for time_step in time_steps:
        clusters = vasculature_dict[time_step]
        for cluster in clusters:
            total_cluster_count += 1
            mesenchymal_count += cluster[0]
            epithelial_count += cluster[1]
            if cluster[0] + cluster[1] > 1:
                multicellular_cluster_count += 1
    df_export = pd.DataFrame({ "Time": [step], "Mesenchymal cells": [mesenchymal_count], "Epithelial cells" :[epithelial_count], "Multicellular clusters": [multicellular_cluster_count], "Total clusters": [total_cluster_count]})
    return df_export

def generate_data(nameOfTheSimulation):
    simulation_path = os.path.join("Simulations", nameOfTheSimulation)
    
    configs_path = os.path.join(simulation_path, "configs.csv")
    metaspread.configs.load_simulation_configs_for_data_generation(configs_path)
    
    print(f'\tAnalyzing data in the folder {simulation_path}\n')

    # Get the Ecm and Mmp2 data filenames 
    ecm_path = os.path.join(simulation_path, "Ecm")
    mmp2_path = os.path.join(simulation_path, "Mmp2")
    ecm_files_name = [f for f in sorted(os.listdir(ecm_path), key=lambda x: int(re.findall(r'\d+(?=step)', x)[0])) if os.path.isfile(os.path.join(ecm_path, f)) and f.endswith(".csv")]
    mmp2_files_name = [f for f in sorted(os.listdir(mmp2_path), key=lambda x: int(re.findall(r'\d+(?=step)', x)[0])) if os.path.isfile(os.path.join(mmp2_path, f)) and f.endswith(".csv")]

    # Get the vasculature data filename
    vasculature_path = os.path.join(simulation_path, "Vasculature")
    vasculature_files_name = [f for f in os.listdir(vasculature_path) if os.path.isfile(os.path.join(vasculature_path, f)) and f.endswith(".json")]

    all_cells_filename = os.path.join(simulation_path, "CellsData.csv")
    if os.path.isfile(all_cells_filename):
        print("Using cells data at:", all_cells_filename)
    else:
        print("No CellsData.csv cell data found in directory:", simulation_path)
        return

    if ecm_files_name:
        ecm_files_path = [os.path.join(ecm_path, p) for p in ecm_files_name]
        print("Using Ecm data in the folder:", ecm_path)
    else:
        print("No .csv Ecm data found in directory:", ecm_path)
        return

    if mmp2_files_name:
        mmp2_files_path = [os.path.join(mmp2_path, p) for p in mmp2_files_name]
        print("Using Mmp2 data in the folder:", mmp2_path)
    else:
        print("No .csv Mmp2 data found in directory:", mmp2_path)
        return

    if vasculature_files_name:
        print("Using vasculature data at:", vasculature_path)
    else:
        print("No .json vasculature data found in directory:", simulation_path)
        return

    step_size = metaspread.configs.data_collection_period
    real_delta_time = 40 * metaspread.configs.th/0.001 #in seconds (the original ratio is 40 seconds/0.001 non-dimensional time)
    grids_number = metaspread.configs.grids_number
    configs_max_step = metaspread.configs.max_steps
    all_cells_dataframe = pd.read_csv(all_cells_filename, converters={"Position": ast.literal_eval})
    all_cells_dataframe = all_cells_dataframe[["Step", "Position", "Phenotype", "Grid", "Agent Type", "Ruptured"]]
    max_step = max(all_cells_dataframe["Step"])
    if configs_max_step >= max_step:
        print(f"Warning: the run for this simulation terminated early")
        print(f"Max step reached is {max_step} while {configs_max_step} was expected.")

    # Path to save the data:
    data_folder = "Data analysis"
    data_path = os.path.join(simulation_path, data_folder)

    tumor_data_path = os.path.join(data_path, "Tumor dynamics")
    cells_data_path = os.path.join(data_path, "Cells growth")
    vasculature_data_path = os.path.join(data_path, "Vasculature dynamics")

    # Create folder for all the data analysis
    os.makedirs(data_path, exist_ok = True)
    os.makedirs(tumor_data_path, exist_ok = True)
    os.makedirs(cells_data_path, exist_ok = True)
    os.makedirs(vasculature_data_path, exist_ok = True)

    print(f"\nSaving tumor data in the folder:", tumor_data_path)
    print("Saving cells numbers data in the folder:", cells_data_path)
    print("Saving vasculature data in the folder:", vasculature_data_path)
    
    for grid_id in range(1, grids_number+1):
        print(f'\nGrid: {grid_id}')

        print(f'\tSaving tumor data...')
        df_radius_diameter_history = pd.DataFrame(columns=['Centroid x', 'Centroid y', 'Radius', 'Diameter', 'Step', 'Grid Id'])
        for id, step in enumerate(range(step_size,max_step+1,step_size)):
            real_time_at_step = real_delta_time * step
            if grid_id == 1:
                # save_cancer(all_cells_dataframe, grid_id, step, real_time_at_step, tumor_data_path)
                (centroid, radius, diameter) = save_cancer(all_cells_dataframe, grid_id, step, real_time_at_step, tumor_data_path)
                new_row = pd.DataFrame({'Centroid x': [centroid[0]], 'Centroid y': [centroid[1]],'Radius': [radius], 'Diameter': [diameter], 'Step': [step], 'Grid Id': [grid_id]})
                df_radius_diameter_history = pd.concat([df_radius_diameter_history, new_row])
            else:
                save_cancer(all_cells_dataframe, grid_id, step, real_time_at_step, tumor_data_path)
        if grid_id == 1:
            path = os.path.join(tumor_data_path, f'Tumor radius and diameter history in grid {grid_id}.csv')
            if not os.path.isfile(path):
                df_radius_diameter_history.to_csv(path)

        print(f'\tSaving cells numbers graph data...')
        df_csv_last_step = pd.DataFrame()
        for id, step in enumerate(range(step_size,max_step+1,step_size)):
            real_time_at_step = real_delta_time * step
            df_csv_last_step = save_growth_data(all_cells_dataframe, grid_id , cells_data_path, step, real_time_at_step, real_delta_time, df_csv_last_step)

    print(f'Saving vasculature...')
    df_export = pd.DataFrame(columns=["Time", "Mesenchymal cells", "Epithelial cells", "Multicellular clusters", "Total clusters"])
    for step in range(step_size,max_step+1,step_size):
        path = os.path.join(vasculature_data_path, f'Vasculature-step{step}.csv')
        if os.path.isfile(path):
            continue
        file_vasculature_on_step = os.path.join(vasculature_path, f"Vasculature-{step}step.json")
        row = get_vasculature_state_at_step(vasculature_data_path, file_vasculature_on_step, step)
        df_export = pd.concat([df_export, row])
        df_export.to_csv(path)

def generate_data_vasculature_only(nameOfTheSimulation):
    simulation_path = os.path.join("Simulations", nameOfTheSimulation)
    
    configs_path = os.path.join(simulation_path, "configs.csv")
    metaspread.configs.load_simulation_configs_for_data_generation

    # Get the vasculature data filename
    vasculature_path = os.path.join(simulation_path, "Vasculature")
    vasculature_files_name = [f for f in os.listdir(vasculature_path) if os.path.isfile(os.path.join(vasculature_path, f)) and f.endswith(".json")]

    all_cells_filename = os.path.join(simulation_path, "CellsData.csv")
    if os.path.isfile(all_cells_filename):
        print("Using cells data at:", all_cells_filename)
    else:
        print("No CellsData.csv cell data found in directory:", simulation_path)
        return
    if vasculature_files_name:
        print("Using vasculature data at:", vasculature_path)
    else:
        print("No .json vasculature data found in directory:", simulation_path)
        return

    print("Loading CellsData.csv. This might take a minute...")
    step_size = metaspread.configs.data_collection_period
    configs_max_step = metaspread.configs.max_steps
    all_cells_dataframe = pd.read_csv(all_cells_filename, converters={"Position": ast.literal_eval})
    all_cells_dataframe = all_cells_dataframe[["Step", "Position", "Phenotype", "Grid", "Agent Type", "Ruptured"]]
    max_step = max(all_cells_dataframe["Step"])
    if configs_max_step >= max_step:
        print(f"Warning: the run for this simulation terminated early")
        print(f"Max step reached is {max_step} while {configs_max_step} was expected.")

    # Path to save the vasculature data:
    data_folder = "Data analysis"
    data_path = os.path.join(simulation_path, data_folder)
    vasculature_data_path = os.path.join(data_path, "Vasculature dynamics")

    print(f'Saving vasculature...')
    df_export = pd.DataFrame(columns=["Time", "Mesenchymal cells", "Epithelial cells", "Multicellular clusters", "Total clusters"])
    for step in range(step_size,max_step+1,step_size):
        file_vasculature_on_step = os.path.join(vasculature_path, f"Vasculature-{step}step.json")
        row = get_vasculature_state_at_step(vasculature_data_path, file_vasculature_on_step, step)
        df_export = pd.concat([df_export, row])
        path = os.path.join(vasculature_data_path, f'Vasculature-step{step}.csv')
        df_export.to_csv(path)

def get_cluster_centroid_radius_and_diameter(ccells_positions, grid_id):
    """"
    Calculates the radius and diameter of the cancer cells in a given site.

    Input:
        model_dataframe: the dataframe containing al information about the model
        grid_id: the grid number for which the radius and diameter will be calculated.
    Returns:
        radius: the maximum of all the distances from each cell to the cell's centroid.
        diameter: the maximum of all the cell-cell distances.
    """
    if ccells_positions.empty or grid_id != 1:
        return ([np.nan, np.nan], np.nan, np.nan)
    ccells_positions= list(ccells_positions['Position'].unique())
    centroid = np.average(ccells_positions, axis=0)
    #calculating radius
    radii  = np.linalg.norm(ccells_positions - centroid, axis=1)
    radius = radii.max()
    #calculating diameter
    dist_matrix = get_distance_matrix(ccells_positions)
    diameter = dist_matrix.max()
    return (centroid, radius, diameter)

def get_distance_matrix(vectors):
    """
    Calculates the distance matrix between all the vectors in a list

    Input:
        vectors (list of numpy 2 by 1 arrays): 
    Returns:
        distance matrix: len(vectors) by len(Vectors) numpy array with the distances
        for all vectors.
    """
    vectors = np.array(vectors)
    x = np.sum(vectors**2,axis=1)
    xx = np.matmul(vectors,vectors.T)
    x2 = x.reshape(-1,1) #transposing the vector
    return np.sqrt(x2-2*xx+x)


import numpy as np
import matplotlib.pyplot as plt
#import random

# Returns an array with 1 in the region that contains the N centre-most grid points and 0 on the others
def find_quasi_circle(N, grid_width, grid_height):
    '''
    Returns the coordinate of the centre-most points and also an
    array with 1 in that region and 0 on the rest
    '''

    current_grid = np.zeros((grid_height, grid_width))
    less_than_ideal_radius = np.sqrt(N/np.pi) / 1.2   
    n_cells_current = 0

    # Center of the grid in 0 based indexing
    a = grid_width/2 - 0.5
    b = grid_height/2 - 0.5

    # Increases the radius of the circle until if finds the correct size
    for radius in np.arange(less_than_ideal_radius, (min(grid_width, grid_height)/2), 0.1):
        last_grid = current_grid.copy()
        n_cells_last = n_cells_current

        # fill the circle
        for x in range(grid_width):
            for y in range(grid_height):
                if (x-a)**2 + (y-b)**2 < radius**2:
                    current_grid[x,y] = 1

        n_cells_current = np.count_nonzero(current_grid == 1)

        if n_cells_current >= N:

            if n_cells_current == N:
                last_grid = current_grid

            if n_cells_current > N:
                n_cells_to_add = N - n_cells_last
                circle_border = np.ma.masked_where(last_grid == 1, current_grid)

                # Fill the outer cells in a diametrically opposit counter-clocksise way
                outer_p = np.where(circle_border == 1)
                coords_outer_p = list(zip(outer_p[0], outer_p[1]))
                sign = 1
                for i in range(n_cells_to_add):
                    coord_to_add = coords_outer_p[i*sign]
                    last_grid[coord_to_add] = 1
                    sign *= -1

            # Create a list with possible places
            possible_places = np.where(last_grid == 1)
            coords = [list(tup) for tup in zip(possible_places[0], possible_places[1], (0 for i in possible_places[0]))]
            return last_grid, coords